import datetime
from ..internals.base_operator import BaseOperator
from ...utils import paths
from ...data.formats import json


class BaseWriterOperator(BaseOperator):

    def __init__(
            self,
            *,
            writer,
            inner_writer,
            **kwargs):
        """
        Writers should extend this class
        """
        super().__init__(**kwargs)
        self.inner_writer = inner_writer(**kwargs)
        self.writer = writer(
                inner_writer=inner_writer,
                **kwargs)

    def execute(self, data: dict = {}, context: dict = {}):
        self.writer.append(data)
        return data, context

    def finalize(self, context: dict = None):

        if isinstance(context, dict):

            # if we have a profiler in the path, it will 
            profile = context.get('mabel:profile')
            if profile:
                timestamp = datetime.datetime.now().strftime('_SYS.as_at_%Y%m%d-%H%M%S')
                profile_path = self.writer.dataset + timestamp + '.profile.json'

                fn_bucket, fn_path, fn_stem, fn_ext = paths.get_parts(self.inner_writer.filename)
                bucket, path, stem, ext = paths.get_parts(profile_path)

                if fn_bucket != bucket:
                    profile_path = path + stem + ext

                self.inner_writer.commit(json.serialize(profile, indent=True, as_bytes=True), profile_path)

        if not context:
            context = {}
            
        if self.writer.schema:
            context['schema_object'] = self.writer.schema

        self.writer.finalize()
        return context

    def __del__(self):
        self.writer.finalize()



import sys
import os.path
import threading
from typing import Callable, Optional, Tuple, List
from .internals.threaded_reader import threaded_reader
from .internals.alpha_processed_reader import processed_reader
from .internals.parsers import pass_thru_parser, block_parser, json_parser, xml_parser
from .internals.filters import Filters
from ..formats.dictset import select_record_fields, select_from
from ..formats.dictset.display import html_table, ascii_table
from ..formats import json
from ...logging import get_logger
from ...errors import InvalidCombinationError, MissingDependencyError
from ...index.index import Index, safe_field_name
from ...utils.parameter_validator import validate


# available parsers
PARSERS = {
    "json": json_parser,
    "text": pass_thru_parser,
    "block": block_parser,
    "pass-thru": pass_thru_parser,
    "xml": xml_parser
}

RULES = [
    {"name": "as_at", "required": False, "warning": "Time Travel (as_at) is Alpha - it's interface may change and some features may not be supported", "incompatible_with": ['start_date', 'end_date']},
    {"name": "cursor", "required": False, "warning": "`cursor` is Alpha - it's interface may change and some features may not be supported", "incompatible_with": ['thread_count', 'fork_processes']},
    {"name": "dataset", "required": True, "warning": None, "incompatible_with": []},
    {"name": "end_date", "required": False, "warning": None, "incompatible_with": []},
    {"name": "extension", "required": False, "warning": None, "incompatible_with": []},
    {"name": "filters", "required": False, "warning": None, "incompatible_with": []},
    {"name": "fork_processes", "required": False, "warning": "Forked Reader is Alpha - it's interface may change and some features may not be supported", "incompatible_with": ['thread_count']},
    {"name": "from_path", "required": False, "warning": "DEPRECATION: Reader \'from_path\' parameter will be replaced with \'dataset\'", "incompatible_with": ['dataset']},
    {"name": "inner_reader", "required": False, "warning" :None, "incompatible_with": []},
    {"name": "project", "required": False, "warning": None, "incompatible_with": []},
    {"name": "raw_path", "required": False, "warning": None, "incompatible_with": []},
    {"name": "row_format", "required": False, "warning": None, "incompatible_with": []},
    {"name": "select", "required": False, "warning": None, "incompatible_with": []},
    {"name": "self", "required": True, "warning": None, "incompatible_with": []},
    {"name": "start_date", "required": False, "warning": None, "incompatible_with": []},
    {"name": "step_back_days", "required": False, "warning": None, "incompatible_with": []},
    {"name": "thread_count", "required": False, "warning": "Threaded Reader is Beta - use in production systems is not recommended", "incompatible_with": []},
    {"name": "where", "required": False, "warning": "`where` will be deprecated, use `filters` or `dictset.select_from` instead", "incompatible_with": ['filters']},
]


class Reader():

    @validate(RULES)
    def __init__(
        self,
        *,  # force all paramters to be keyworded
        select: list = ['*'],
        dataset: str = None,
        where: Optional[Callable] = None,
        filters: Optional[List[Tuple[str, str, object]]] = None,
        inner_reader=None,   # type:ignore
        row_format: str = "json",
        **kwargs):
        """
        Reads records from a data store, opinionated toward Google Cloud Storage but a
        filesystem reader is available to assist with local development.

        The Reader will iterate over a set of files and return them to the caller as a
        single stream of records. The files can be read from a single folder or can be
        matched over a set of date/time formatted folder names. This is useful to read
        over a set of logs. The date range is provided as part of the call; this is
        essentially a way to partition the data by date/time.

        The reader can filter records to return a subset, for JSON formatted data the
        records can be converted to dictionaries before filtering. JSON data can also
        be used to select columns, so not all read data is returned.

        The reader does not support aggregations, calculations or grouping of data, it
        is a log reader and returns log entries. The reader can convert a set into 
        _Pandas_ dataframe, or the _dictset_ helper library can perform some activities
        on the set in a more memory efficient manner.

        Note:
            Different _inner_readers_ may take or require additional parameters. This
            class has a decorator which helps to ensure it is called correctly.

        Parameters:
            select: list of strings (optional):
                A list of the names of the columns to return from the dataset, the
                default is all columns
            dataset: string:
                The path to the data
            where: callable (optional):
                **TO BE DEPRECATED**
                A method (function or lambda expression) to filter the returned
                records, where the function returns True the record is returned, False
                the record is skipped. The default is all records
            filters: List of tuples (optional)
                Rows which do not match the filter predicate will be removed from
                scanned data. Default is no filtering.
                Each tuple has format: (`key`, `op`, `value`) and compares the key with
                the value. The supported op are: `=` or `==`, `!=`,  `<`, `>`, `<=`,
                `>=`, `in`, `!in` (not in), `contains` and `!contains` (doesn't 
                contain) and `like`. If the `op` is `in` or `!in`, the `value` must be
                a _list_. `like` performs similar to the SQL operator `%` is a
                multicharacter wildcard and `_` is a single character wildcard.
                If a field is indexed, it will be used only for '==' and 'in' and
                'contains' operations.
            inner_reader: BaseReader (optional):
                The reader class to perform the data access Operators, the default is
                GoogleCloudStorageReader
            row_format: string (optional):
                Controls how the data is interpretted. 'json' will parse to a
                dictionary before _select_ or _where_ is applied, 'text' will just
                return the line that has been read, 'block' will return the content of
                a file as a record. the default is 'json'.
            start_date: datetime (optional):
                The starting date of the range to read over - if used with
                _date_range_, this value will be preferred, default is today
            end_date: datetime (optional):
                The end date of the range to read over - if used with _date_range_,
                this value will be preferred, default is today
            thread_count: integer (optional):
                **BETA**
                Use multiple threads to read data files, the default is to not use
                additional threads, the maximum number of threads is 8
            step_back_days: integer (optional):
                The number of days to look back if data for the current date is not
                available.
            fork_processes: boolean (alpha):
                **ALPHA**
                Create parallel processes to read data files
            as_at: datetime (alpha)
                **ALPHA**
                Time travel
            cursor: dictionary (alpha)
                **ALPHA**
                Resume read from a given point (assumes other parameters are the same).
                If a JSON string is provided, it will converted to a dictionary.

        Yields:
            dictionary (string if data format is 'text')

        Raises:
            TypeError
                Reader _select_ parameter must be a list
            TypeError
                Reader _where_ parameter must be Callable or None
            TypeError
                Data format unsupported
            InvalidCombinationError
                Forking and Threading can not be used at the same time
        """
        if not isinstance(select, list):  # pragma: no cover
            raise TypeError("Reader 'select' parameter must be a list")
        if where is not None and not hasattr(where, '__call__'):
            raise TypeError("Reader 'where' parameter must be Callable or None")

        # load the line converter
        self._parse = PARSERS.get(row_format.lower())
        if self._parse is None:  # pragma: no cover
            raise TypeError(F"Row format unsupported: {row_format} - valid options are {list(PARSERS.keys())}.")

        if dataset is None and kwargs.get('from_path'):  # pragma: no cover
            dataset = kwargs['from_path']

        # lazy loading of dependency
        if inner_reader is None:
            from ...adapters.google import GoogleCloudStorageReader
            inner_reader = GoogleCloudStorageReader
        # instantiate the injected reader class
        self.reader_class = inner_reader(dataset=dataset, **kwargs)  # type:ignore

        self.cursor = kwargs.get('cursor', None)
        if isinstance(self.cursor, str):
            self.cursor = json.parse(self.cursor)
        if not isinstance(self.cursor, dict):
            self.cursor = {}

        self.select = select.copy()
        self.where: Optional[Callable] = where

        # initialize the reader
        self._inner_line_reader = None

        arg_dict = kwargs.copy()
        arg_dict['select'] = F'{select}'
        arg_dict['dataset'] = F'{dataset}'
        arg_dict['inner_reader'] = F'{inner_reader.__name__}'  # type:ignore
        arg_dict['row_format'] = F'{row_format}'
        get_logger().debug(json.serialize(arg_dict))

        # number of days to walk backwards to find records
        self.step_back_days = int(kwargs.get('step_back_days', -1))
        if self.step_back_days > 0 and self.reader_class.start_date != self.reader_class.end_date:  # pragma: no cover
            raise InvalidCombinationError("step_back_days can only be used when the start and end dates are the same")

        if row_format != 'pass-thru' and kwargs.get('extension') == '.parquet':  # pragma: no cover
            raise InvalidCombinationError("`parquet` extension much be used with the `pass-thru` row_format")

        self.filters = None
        self.indexable_fields = []
        if filters:
            self.filters = Filters(filters)   
            self.indexable_fields = self._get_indexable_filter_columns(self.filters.predicates)

        """ FEATURES IN DEVELOPMENT """

        # threaded reader
        self.thread_count = int(kwargs.get('thread_count', 0))

        # multiprocessed reader
        self.fork_processes = bool(kwargs.get('fork_processes', False))

        # time travel
        self.as_at = kwargs.get('as_at')

    """
    Iterable

    Use this class as an iterable:

        for line in Reader("file"):
            print(line)
    """
    def _get_indexable_filter_columns(self, predicate):
        """
        Returns all of the columns in a filter which the operation benefits
        from an index

        This creates an list of tuples of (field,value) that we can feed to the
        index search.
        """
        INDEXABLE_OPS = {'=', '==', 'is', 'in', 'contains'}
        if predicate is None:
            return []
        if isinstance(predicate, tuple):
            key, op, value = predicate
            if op in INDEXABLE_OPS:
                return [(key, value,)]
        if isinstance(predicate, list):
            if all([isinstance(p, tuple) for p in predicate]):
                return [(k,v,) for k,o,v in predicate if o in INDEXABLE_OPS]
            if all([isinstance(p, list) for p in predicate]):
                columns = []
                for p in predicate:
                    columns += self._get_indexable_filter_columns(p)
                return columns
        return []    # pragma: no cover

    def _is_system_file(self, filename):
        if '_SYS.' in filename:
            base = os.path.basename(filename)
            return base.startswith('_SYS.')
        return False

    def _read_blob(self, blob, blob_list):
        """
        This wraps the blob reader, including the filters and indexers
        """
        get_logger().debug(F"Reading from `{blob}`, thread: {threading.get_ident()}")
        # If an index exists, get the rows we're interested in from the index
        rows = None
        for field, filter_value in self.indexable_fields:
            # does an index file for this file and record exist
            path, file = os.path.split(blob)
            stem, ext = os.path.splitext(file)
            index_file = path + '/_SYS.' + stem + '.' + safe_field_name(field) + '.index'

            if index_file in blob_list:
                get_logger().debug(F"Reading index from blob `{index_file}`")
                index_stream = self.reader_class.get_blob_stream(index_file)
                index = Index(index_stream)
                rows = rows or []
                rows += index.search(filter_value)

        # read the rows from the file
        ds = self.reader_class.get_records(blob, rows)
        # reformat the rows
        ds = self._parse(ds)
        # filter the rows, either with the filters or `dictset.select_from`
        if self.filters:
            yield from self.filters.filter_dictset(ds)
        else:
            yield from select_from(ds, where=self.where)

    def _create_line_reader(self):
        blob_list = self.reader_class.get_list_of_blobs()

        # handle stepping back if the option is set
        if self.step_back_days > 0:
            while not bool(blob_list) and self.step_back_days >= self.reader_class.days_stepped_back:
                self.reader_class.step_back_a_day()
                blob_list = self.reader_class.get_list_of_blobs()
            if self.step_back_days < self.reader_class.days_stepped_back:
                get_logger().alert(F"No data found in last {self.step_back_days} days - aborting")
                sys.exit(-1) 
            if self.reader_class.days_stepped_back > 0:
                get_logger().warning(F"Stepped back {self.reader_class.days_stepped_back} days to {self.reader_class.start_date} to find last data, my limit is {self.step_back_days} days.")

        readable_blobs = [b for b in blob_list if not self._is_system_file(b)]

        # skip to the blob in the cursor
        if self.cursor.get('blob'):
            skipped = 0
            while len(readable_blobs) > 0:
                if readable_blobs[0] != self.cursor.get('blob'):
                    readable_blobs.pop(0)
                    skipped += 1
                else:
                    get_logger().debug(F"Reader found {len(readable_blobs)} sources to read data from after {skipped} jumped to get to cursor.")
                    break
        else:
            get_logger().debug(F"Reader found {len(readable_blobs)} sources to read data from.")
        
        if self.thread_count > 0:
            yield from threaded_reader(readable_blobs, blob_list, self)

        elif self.fork_processes:
            yield from processed_reader(readable_blobs, self.reader_class, self._parse, self.where)

        else:
            base_offset = self.cursor.get('offset', 0)
            for blob in readable_blobs:
                self.cursor['blob'] = blob
                self.cursor['offset'] = -1
                local_reader = self._read_blob(blob, blob_list)
                for burn in range(base_offset):
                    next(local_reader, None)
                for offset, record in enumerate(local_reader):
                    self.cursor['offset'] = offset + base_offset
                    yield record
                base_offset = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self._inner_line_reader is None:
            self._inner_line_reader = self._create_line_reader()

        # get the the next line from the reader
        record = self._inner_line_reader.__next__()
        if self.select != ['*']:
            record = select_record_fields(record, self.select)
        return record


    """
    Context Manager

    Use this class using the 'with' statement:

        with Reader("file") as r:
            line = r.read_line()
            while line:
                print(line)
    """

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass  # exist needs to exist to be a context manager

    def read_line(self):
        """
        Read the next line from the _Reader_.

        Returns:
            dictionary (or string)
        """
        try:
            return self.__next__()
        except StopIteration:
            return None

    """
    Exports
    """
    def to_pandas(self):
        """
        Load the contents of the _Reader_ to a _Pandas_ DataFrame.

        Returns:
            Pandas DataFrame
        """
        try:
            import pandas as pd  # type:ignore
        except ImportError:  # pragma: no cover
            raise MissingDependencyError("`pands` is missing, pleaseinstall or include in requirements.txt")
        return pd.DataFrame(self)

    def __repr__(self):  # pragma: no cover

        def is_running_from_ipython():
            """
            True when running in Jupyter
            """
            try:
                from IPython import get_ipython  # type:ignore
                return get_ipython() is not None
            except Exception:
                return False

        if is_running_from_ipython():
            from IPython.display import HTML, display  # type:ignore
            html = html_table(self, 5)
            display(HTML(html))
            return ''  # __repr__ must return something
        else:
            return ascii_table(self, 5)



from typing import Iterator, Iterable, List


def html_table(
        dictset: Iterator[dict],
        limit: int = 5):
    """
    Render the dictset as a HTML table.

    NOTE:
        This exhausts generators so is only recommended to be used on lists.

    Parameters:
        dictset: iterable of dictionaries
            The dictset to render
        limit: integer (optional)
            The maximum number of record to show in the table, defaults to 5

    Returns:
        string (HTML table)
    """
    def _to_html_table(data, columns):

        yield '<table class="table table-sm">'
        for counter, record in enumerate(data):
            if counter == 0:
                yield '<thead class="thead-light"><tr>'
                for column in columns:
                    yield '<th>' + column + '<th>\n'
                yield '</tr></thead><tbody>'

            if (counter % 2) == 0:
                yield '<tr style="background-color:#F4F4F4">'
            else:
                yield '<tr>'
            for column in columns:
                yield '<td>' + str(record.get(column)) + '<td>\n'
            yield '</tr>'

        yield '</tbody></table>'

    rows = []
    columns = []  # type:ignore
    for i, row in enumerate(dictset):
        rows.append(row)
        columns = columns + list(row.keys())
        if i == limit:
            break
    columns = set(columns)  # type:ignore

    import types
    footer = ''
    if isinstance(dictset, types.GeneratorType):
        footer = f'\n<p>top {limit} rows x {len(columns)} columns</p>'
        footer += '\nNOTE: the displayed records have been spent'
    if isinstance(dictset, list):
        footer = f'\n<p>{len(dictset)} rows x {len(columns)} columns</p>'

    return ''.join(_to_html_table(rows, columns)) + footer


def ascii_table(
        dictset: Iterator[dict],
        limit: int = 5):
    """
    Render the dictset as a ASCII table.

    NOTE:
        This exhausts generators so is only recommended to be used on lists.

    Parameters:
        dictset: iterable of dictionaries
            The dictset to render
        limit: integer (optional)
            The maximum number of record to show in the table, defaults to 5

    Returns:
        string (ASCII table)
    """
    result = []
    columns: dict = {}
    cache = []

    # inspect values
    for count, row in enumerate(dictset):
        if count == limit:
            break

        cache.append(row)
        for k, v in row.items():
            length = max(len(str(v)), len(k))
            if length > columns.get(k, 0):
                columns[k] = length

    # draw table
    bars = []
    for header, width in columns.items():
        bars.append('─' * (width + 2))

    # display headers
    result.append('┌' + '┬'.join(bars) + '┐')
    result.append('│' + '│'.join([k.center(v + 2) for k, v in columns.items()]) + '│')
    result.append('├' + '┼'.join(bars) + '┤')

    # display values
    for row in cache:
        result.append('│' + '│'.join([str(v).center(columns[k] + 2) for k, v in row.items()]) + '│')

    # display footer
    result.append('└' + '┴'.join(bars) + '┘')

    return '\n'.join(result)


BAR_CHARS = [r" ", r"▁", r"▂", r"▃", r"▄", r"▅", r"▆", r"▇", r"█"]

def draw_histogram_bins(bins: List[int]):
    """
    Draws a pre-binned set off histogram data
    """
    mx = max(bins)
    bar_height = (mx / 8)
    if bar_height == 0:
        return ' ' * len(bins)
    
    histogram = ''
    for value in bins:
        if value == 0:
            histogram += BAR_CHARS[0]
        else:
            height = int(value / bar_height)
            histogram += BAR_CHARS[height]
        
    return histogram

def histogram(values: Iterable[int], number_of_bins: int = 10):
    """
    """
    bins = [0] * number_of_bins
    mn = min(values)
    mx = max(values)
    interval = ((1 + mx) - mn) / (number_of_bins - 1)

    for v in values:
        bins[int(v / interval)] += 1
    return draw_histogram_bins(bins)



"""
MinIO Bin Writer

May support AWS S3 - untested
"""
import time
import io
from .base_bin import BaseBin
from ...errors import MissingDependencyError
try:
    from minio import Minio  # type:ignore
    minio_installed = True
except ImportError:  # pragma: no cover
    minio_installed = False


class MinioBin(BaseBin):

    def __init__(
            self,
            bin_name: str,
            end_point: str,
            bucket: str,
            path: str,
            access_key: str,
            secret_key: str,
            secure: bool = True):

        if not minio_installed:  # pragma: no cover
            raise MissingDependencyError("`minio` is missing, please install or include in requirements.txt")

        self.client = Minio(end_point, access_key, secret_key, secure=secure)
        self.bucket = bucket
        self.path = path
        self.name = bin_name

    def __str__(self) -> str:
        return self.name

    def __call__(
            self,
            record: str,
            id_: str = ''):

        filename = F"{self.path}/{self._date_part()}/{id_}{time.time_ns()}.txt"
        record_bytes = record.encode('utf-8')
        record_stream = io.BytesIO(record_bytes)

        self.client.put_object(
                self.bucket,
                filename,
                record_stream,
                length=len(record_bytes))

        return filename



"""
Bit Array

(C) 2021 Justin Joyce.

This is a limited drop-in replacement for the bitarry library which may not
work on all environments due to needing a C compiler to be present.

If bitarry is not installed, this will take it's place, but it is only
a partial implementation and is many times slower.


Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class bitarray:
    def __init__(self, size: int = 0, endian: str = ""):
        self.size = size
        self.bits = [0 for i in range((size + 7) >> 3)]

    def setall(self, value=0):
        for i in range(self.size):
            self.__setitem__(i, value)

    def __setitem__(self, bit, value):
        b = self.bits[bit >> 3]
        if value:
            self.bits[bit >> 3] = b | 1 << (bit % 8)
        else:
            self.bits[bit >> 3] = b & ~(1 << (bit % 8))

    def __getitem__(self, bit):
        b = self.bits[bit >> 3]
        return (b >> (bit % 8)) & 1

    def __repr__(self):
        def _inner():
            for i in range(self.size):
                yield str(self[i])

        return f"bitarray('{''.join(list(_inner()))}')"

    def tofile(self, filehandle):
        filehandle.write(bytes(self.bits))

    def fromfile(self, filehandle):
        self.size = 0
        b = filehandle.read(1)
        while b:
            i = int.from_bytes(b, byteorder="big")
            self.size += 8
            self.bits.append(i)
            b = filehandle.read(1)



from ...flows.internals.base_writer_operator import BaseWriterOperator
from ...data import BatchWriter
from ...adapters.minio import MinIoWriter


class MinIoBatchWriterOperator(BaseWriterOperator):

    def __init__(
            self,
            *,
            end_point,
            access_key,
            secret_key,
            secure: bool = False,
            **kwargs):
        """
        """
        kwargs['end_point'] = end_point
        kwargs['access_key'] = access_key
        kwargs['secret_key'] = secret_key
        kwargs['secure'] = secure

        super().__init__(
                writer=BatchWriter,
                inner_writer=MinIoWriter,
                **kwargs)




# google-cloud-logging
# pydantic
import os
import logging
import datetime
from typing import Union, Optional
from ..utils import is_running_from_ipython
try:
    from google.cloud import logging as stackdriver  # type:ignore
    from google.cloud.logging import DESCENDING  # type:ignore
except ImportError:
    stackdriver = None  # type:ignore


class GoogleLogger():

    @staticmethod
    def supported():
        use_logger = []
        use_logger.append(stackdriver is not None)
        use_logger.append(not is_running_from_ipython())
        use_logger.append(not os.environ.get('IGNORE_STACKDRIVER', False))
        return all(use_logger)

    @staticmethod
    def write_event(
            message: Union[str,dict],
            system: Optional[str] = None,
            severity: Optional[int] = logging.DEBUG):

        from .create_logger import LOG_NAME
        from .levels import LEVELS_TO_STRING, LEVELS

        print(f"{LOG_NAME}| {LEVELS_TO_STRING.get(severity, 'UNKNOWN')} | {datetime.datetime.now().isoformat()} | {message}")  # type:ignore

        # rewrite one of the levels
        LEVELS_TO_STRING[LEVELS.AUDIT] = "NOTICE"

        client = stackdriver.Client()
        logger = client.logger(LOG_NAME)
        
        labels = {}
        if system:
            labels['system'] = system

        if isinstance(message, dict):
            logger.log_struct(
                    info=message,
                    severity=LEVELS_TO_STRING.get(severity),  # type:ignore
                    labels=labels)
        else:
            logger.log_text(
                    text=F"{message}",
                    severity=LEVELS_TO_STRING.get(severity),  # type:ignore
                    labels=labels)

    def __init__(self):
        from .levels import LEVELS

        self.level = int(os.environ.get("LOGGING_LEVEL", 25))
        self.debug = self.create_logger(LEVELS.DEBUG)
        self.info = self.create_logger(LEVELS.INFO)
        self.warning = self.create_logger(LEVELS.WARNING)
        self.error = self.create_logger(LEVELS.ERROR)
        self.audit = self.create_logger(LEVELS.AUDIT)
        self.alert = self.create_logger(LEVELS.ALERT)


    def create_logger(self, level):
        from .create_logger import LOG_NAME

        def base_logger(message):
            GoogleLogger.write_event(
                message=message,
                system=LOG_NAME,
                severity=level)

        def do_nothing(message):
            pass

        if level > self.level:
            return base_logger
        else:
            return do_nothing

    def setLevel(self, level):
        self.level = level
        self.debug = self.create_logger(LEVELS.INFO)
        self.info = self.create_logger(LEVELS.INFO)
        self.warning = self.create_logger(LEVELS.WARNING)
        self.error = self.create_logger(LEVELS.ERROR)
        self.audit = self.create_logger(LEVELS.AUDIT)
        self.alert = self.create_logger(LEVELS.ALERT)




import os
import json
import threading
from time import sleep
from ..logging import get_logger

can_use_resource_lib = True
try:
    import resource
except ImportError:
    can_use_resource_lib = False
    import psutil  # type:ignore


class ResourceMonitor:

    slots = "frequency"

    def __init__(self, frequency=1):
        self.frequency = frequency
        if not can_use_resource_lib:
            self.process = psutil.Process(os.getpid())

    def resource_usage(self):
        while True:
            if can_use_resource_lib:
                memory_usage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
            else:
                memory_usage = self.process.memory_info()[0]
            get_logger().info(json.dumps({"memory": memory_usage}))
            sleep(self.frequency)


monitor = ResourceMonitor()
resource_thread = threading.Thread(target=monitor.resource_usage)
resource_thread.name = "mabel-resource-monitor"
resource_thread.daemon = True
resource_thread.start()





COLORS = {
    "{OFF}": "\033[0m",  # Text Reset
    # Regular Colors
    "{BLACK}": "\033[0;30m",  # Black
    "{RED}": "\033[0;31m",  # Red
    "{GREEN}": "\033[0;32m",  # Green
    "{YELLOW}": "\033[0;33m",  # Yellow
    "{BLUE}": "\033[0;34m",  # Blue
    "{PURPLE}": "\033[0;35m",  # Purple
    "{CYAN}": "\033[0;36m",  # Cyan
    "{WHITE}": "\033[0;37m",  # White
    # Bold
    "{BOLD_BLACK}": "\033[1;30m",  # Black
    "{BOLD_RED}": "\033[1;31m",  # Red
    "{BOLD_GREEN}": "\033[1;32m",  # Green
    "{BOLD_YELLOW}": "\033[1;33m",  # Yellow
    "{BOLD_BLUE}": "\033[1;34m",  # Blue
    "{BOLD_PURPLE}": "\033[1;35m",  # Purple
    "{BOLD_CYAN}": "\033[1;36m",  # Cyan
    "{BOLD_WHITE}": "\033[1;37m",  # White
    # Underline
    "{UNDERLINE_BLACK}": "\033[4;30m",  # Black
    "{UNDERLINE_RED}": "\033[4;31m",  # Red
    "{UNDERLINE_GREEN}": "\033[4;32m",  # Green
    "{UNDERLINE_YELLOW}": "\033[4;33m",  # Yellow
    "{UNDERLINE_BLUE}": "\033[4;34m",  # Blue
    "{UNDERLINE_PURPLE}": "\033[4;35m",  # Purple
    "{UNDERLINE_CYAN}": "\033[4;36m",  # Cyan
    "{UNDERLINE_WHITE}": "\033[4;37m",  # White
}


def colorize(record, can_colorize=True):
    if can_colorize:
        for k, v in COLORS.items():
            record = record.replace(k, v)
    else:
        for k, v in COLORS.items():
            record = record.replace(k, "")
    return record


from .base_operator import BaseOperator


class DecoratedOperator(BaseOperator):

    def __init__(self, func=None):
        super().__init__()
        self.name = F"DecoratorOperator:{func.__name__}"
        self.func = func

    def execute(self, data, context):
        response = self.func(data)
        return response, context

def operator(func):
    operator = DecoratedOperator(func=func)
    return operator



from ...formats.json import parse
from ....errors import MissingDependencyError
try:
    import xmltodict  # type:ignore
except ImportError:
    xmltodict = None


def json_parser(ds):
    """ parse each line in the file to a dictionary """
    yield from map(parse, ds)


def pass_thru_parser(ds):
    """ just pass it through """
    yield from ds


def block_parser(ds):
    """ each blob is read as a block """
    if isinstance(ds, str):
        return ds
    yield '\n'.join([r for r in ds])    # pragma: no cover


def xml_parser(ds):
    if not xmltodict:    # pragma: no cover
        raise MissingDependencyError("`xmltodict` is missing, please install or include in requirements.txt")
    yield from map(xmltodict.parse, ds)


"""
Google Cloud Storage Bin Writer
"""
import time
from .base_bin import BaseBin
from ...errors import MissingDependencyError
try:
    from google.cloud import storage  # type:ignore
    google_cloud_storage_installed = True
except ImportError:  # pragma: no cover
    google_cloud_storage_installed = False


class GoogleCloudStorageBin(BaseBin):

    def __init__(
            self,
            bin_name: str,
            project: str,
            bucket: str,
            path: str):

        if not google_cloud_storage_installed:  # pragma: no cover
            raise MissingDependencyError("`google-cloud-storage` is missing, please install or include in requirements.txt")

        client = storage.Client(project=project)
        self.bucket = client.get_bucket(bucket)
        self.path = path
        self.name = bin_name

    def __str__(self) -> str:
        return self.name

    def __call__(
            self,
            record: str,
            id_: str = ''):

        filename = F"{self.path}/{self._date_part()}/{id_}{time.time_ns()}.txt"
        blob = self.bucket.blob(filename)
        blob.upload_from_string(str(record).encode('utf-8'))
        return filename



"""
Diablo: Python Graph Library

(C) 2021 Justin Joyce.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from pathlib import Path
from ..json import serialize
from ....errors import MissingDependencyError

class Graph(object):
    """
    Graph object, optimized for traversal.

    Edges are stored in a dictionary, the key is the source node to speed up 
    finding outgoing edges. The Edges only have three pieces of data:
        - the source node (the key)
        - the target node
        - the relationship
    The target and the relationship are stored as a tuple, the edge dictionary
    stores lists of tuples.

    Nodes are stored as a B+Tree, this gives slightly slower performance 
    than a dictionary but has a distinct advantage in that it sorts the
    values enabling binary searching of the dataset without loading into
    memory.
    """
    __slots__ = ('_nodes', '_edges')

    def __init__(self):
        """
        Directed Graph.
        """
        self._nodes = {}
        self._edges = {}
    

    def _make_a_list(
            self,
            obj):
        """ internal helper method """
        if isinstance(obj, list):
            return obj
        return [obj]


    def save(
            self,
            graph_path):
        """
        Persist a graph to storage. It saves nodes and edges to separate files.

        Parameters:
            graph_path: string
                The folder ?to save the node and edge files to
        """
        path = Path(graph_path)
        path.mkdir(exist_ok=True)

        with open(path / 'edges.jsonl', 'w', encoding='utf8') as edge_file:
            for source, target, relationship in self.edges():
                edge_record = {"source": source, "target": target, "relationship": relationship}
                edge_file.write(serialize(edge_record) + '\n')
        with open(path / 'nodes.jsonl', 'w', encoding='utf8') as node_file:
            for nid, attributes in self.nodes(data=True):
                node_record = {"nid": nid, "attributes": attributes}
                node_file.write(serialize(node_record) + '\n')


    def add_edge(
            self,
            source: str,
            target: str,
            relationship: str):
        """
        Add edge to the graph

        Note:
            This does not create nodes if they don't already exist

        Parameters:
            source: string
                The source node
            target: string
                The target node
            relationship: string
                The relationship between the source and target nodes
        """
        if source not in self._edges:
            targets = []
        else:
            targets = self._edges[source]
        targets.append((target, relationship,))
        self._edges[source] = list(set(targets))


    def add_node(
            self, 
            nid: str,
            attributes:dict = {}):
        """
        Add node to the graph

        Parameters:
            nid: string
                The Node ID for the node (unique)
            attributes: dictionary (optional)
                The attributes of the node
        """
        self._nodes[nid] = attributes


    def nodes(
            self,
            data=False):
        """
        The nodes which comprise the graph

        Parameters:
            data: boolean (optional)
                if True return the details of the nodes, if False just return
                the list of node IDs

        Returns:
            List
        """
        if data:
            return self._nodes.items()
        return list(self._nodes.keys())


    def edges(self):
        """
        The edges which comprise the graph

        Returns:
            Generator of Tuples of (Source, Target and Relationship)
        """
        for s, records in self._edges.items():
            for t, r in records:
                yield s, t, r


    def breadth_first_search(
            self,
            source:str,
            depth:int = 100):
        """
        Search a tree for nodes we can walk to from a given node.

        Parameters:
            source: string
                The node to walk from
            depth: integer
                The maximum distance to walk from source
        Returns:
        """
        # This uses a variation of the algorith used by NetworkX optimized for
        # the Diablo data structures.
        #
        # https://networkx.org/documentation/networkx-1.10/_modules/networkx/algorithms/traversal/breadth_first_search.html#bfs_tree
        
        from collections import deque
        
        distance = 0

        visited = set([source])
        queue = deque([(source, distance, self.outgoing_edges(source),)])

        new_edges = []

        while queue:
            parent, node_distance, children = queue[0]
            if node_distance < depth:
                for child in children:
                    s,t,r = child
                    new_edges.append(child)
                    if t not in visited:
                        visited.add(t)
                        queue.append((t, node_distance + 1, self.outgoing_edges(t),))
            queue.popleft()
        return new_edges

    def outgoing_edges(
            self,
            source):
        """
        Get the list of edges traversable from a given node.

        Parameters:
            source: string
                The node to get the outgoing edges for

        Returns:
            Set of Tuples (Source, Target, Relationship)
        """
        targets = self._edges.get(source) or {}
        return {(source, t, r) for t, r in targets}


    def copy(self):
        g = Graph()
        g._nodes = self._nodes.copy()
        g._edges = self._edges.copy()
        return g


    def to_networkx(self):
        """
        Convert a Diablo graph to a NetworkX graph
        """
        try:
            import networkx as nx  # type:ignore
        except ImportError:  # pragma: no cover
            raise MissingDependencyError('`networx` is missing, please install or include in requirements.txt')
            
        g = nx.DiGraph()
        for s, t, r in self.edges():
            g.add_edge(s, t, relationship=r)
        for node, attribs in self.nodes(True):
            g.add_node(node, **attribs)
        return g

    def epitomize(self):
        """
        Summarize a Graph by reducing to only the node_types and relationships
        """
        g = Graph()
        for s, t, r in self.edges():
            node1 = self[s]
            node2 = self[t]
            if node1 and node2:
                g.add_edge(node1.get('node_type'), node2.get('node_type'), r)
            if node1:
                g.add_node(node1.get('node_type'), {"node_type":node1.get('node_type')})
            if node2:
                g.add_node(node2.get('node_type'), {"node_type":node2.get('node_type')})
        return g


    def __repr__(self):
        return F"Graph - {len(list(self.nodes()))} nodes, {len(list(self.edges()))} edges"

    def __len__(self):
        return len(list(self.nodes()))

    def __getitem__(self, nid):
        return self._nodes.get(nid, {})





#no-maintain-checks
from moz_sql_parser import parse  # type:ignore 
from ..reader import Reader
from ....logging import get_logger
from ...formats import json
from functools import lru_cache
import re
import os

# https://codereview.stackexchange.com/a/248421
_special_regex_chars = {
    ch : '\\'+ch
    for ch in '.^$*+?{}[]|()\\'
}

@lru_cache(4)
def _sql_like_fragment_to_regex(fragment):
    # https://codereview.stackexchange.com/a/36864/229677
    safe_fragment = ''.join([_special_regex_chars.get(ch, ch) for ch in fragment])
    return re.compile('^' + safe_fragment.replace('%', '.*?').replace('_', '.') + '$')

def _eq(x,y):   return x == y
def _neq(x,y):  return x != y
def _lt(x,y):   return x < y
def _gt(x,y):   return x > y
def _lte(x,y):  return x <= y
def _gte(x,y):  return x >= y
def _and(x,y):  return x and y
def _or(x,y):   return x or y
def _add(x,y):  return x + y
def _mul(x,y):  return x * y
def _div(x,y):  return x / y
def _sub(x,y):  return x - y
def _like(x,y): return _sql_like_fragment_to_regex(y).match(str(x))

# functions which implement the Operators
OPERATORS = {
    'eq':   _eq,
    'neq':  _neq,
    'lt':   _lt,
    'gt':   _gt,
    'lte':  _lte,
    'gte':  _gte,
    'and':  _and,
    'or':   _or,
    'like': _like,
    'add':  _add,
    'mul':  _mul,
    'div':  _div,
    'sub':  _sub
}

def _get_operand(operand, row):
    if not isinstance(operand, dict):
        # strings are field names
        if isinstance(operand, str):
            return row[operand]
        # otherwise it's a constant
        return operand
    # string constants are 'literals'
    if 'literal' in operand:
        return operand['literal']
    # some values are handled unusually
    value = [row.get(k) for k, v in operand.items() if k in ['timestamp', 'text']]
    if len(value):
        return value.pop()
    # if we're here, the operand is probably a function
    return _build_where_function(operand)(row)


def _build_where_function(where_object):
    # None is None
    if where_object is None:
        return None
    # the Operator is the head of the dictionary
    operator = list(where_object.keys())[0]
    # the operands are the values in for the key
    operands = where_object.get(operator)
    # if the operands are a list, it's the left and right operands
    if isinstance(operands, list):
        left_operand, right_operand = operands[0], operands[1]
    # if it's a dict, run the Operator against it
    elif isinstance(operands, dict):
        return lambda row: _get_operand(operands, row)

    # we build a function, load some of the values and pass it to the Reader
    return SqlOperator(
            operator,
            left_operand,
            right_operand)


class SqlOperator():
    """
    This is implemented as a class so it can be picked so it can be used with
    the multi-processing library via the fork_processes option in the Reader.
    """
    __slots__ = ('left_operand', 'right_operand', 'operation', '__name__')

    def __init__(
            self,
            operator,
            left_operand,
            right_operand):
        self.left_operand = left_operand
        self.right_operand = right_operand
        self.operation = OPERATORS.get(operator)
        self.__name__ = F"SQL_WHERE('{operator}')"

    def __call__(self, row):
        _left_operand = _get_operand(self.left_operand, row)
        _right_operand = _get_operand(self.right_operand, row)
        return self.operation(_left_operand, _right_operand)


class SqlReader():

    def __init__(self, 
            sql_statement: str,
            **kwargs):
        """
        THIS IS ALPHA - interface and features subject to change

        Use basic SQL queries to filter Reader.

        Parameters:
            sql_statement: string
            kwargs: parameters to pass to the Reader
        
        Note:
            `select` is taken from SQL SELECT
            `dataset` is taken from SQL FROM
            `where` is taken from SQL WHERE
        """
        get_logger().warning("SQL Reader is Alpha - interface and features subject to change")

        sql = parse(sql_statement)
        get_logger().debug(json.serialize(sql))

        field_list = sql.get('select')
        fields = []
        if isinstance(field_list, str):
            if field_list == '*':
                field_list = {'value': '*'}
            field_list = [field_list]
        elif not isinstance(field_list, list):
            field_list = [field_list]
        fields = [field.get('value') for field in field_list]

        table = sql.get('from').replace('.', '/')

        where_statement = _build_where_function(sql.get('where'))
        
        thread_count = 4
        
        self.reader = Reader(
                thread_count=thread_count,
                select=fields,
                where=where_statement,
                dataset=table,
                **kwargs)

    def __iter__(self):
        return self

    def __next__(self):
        # deepcode ignore unguarded~next~call: error should bubble
        return next(self.reader)



"""
No Operation Operator

This does nothing - without error, useful for testing.

It optionally prints the class name - intended to determine when and that
the Operator was run - for testing.
"""
from ..flows.internals.base_operator import BaseOperator


class NoOpOperator(BaseOperator):

    def __init__(self, print_message=False):
        self.print_message = print_message
        super().__init__()

    def execute(self, data={}, context={}):
        if self.print_message:
            print(self.__class__.__name__)
        return data, context



from functools import lru_cache

@lru_cache(1)
def is_running_from_ipython():
    """
    True when running in Jupyter
    """
    try:
        from IPython import get_ipython  # type:ignore

        return get_ipython() is not None
    except Exception:
        return False


import io
from ...data.readers.internals.base_inner_reader import BaseInnerReader


class DiskReader(BaseInnerReader):

    def __init__(self, **kwargs):
        """
        File System Reader
        """
        super().__init__(**kwargs)

    def get_blobs_at_path(self, path):
        if path.exists(): 
            blobs = path.rglob("*")
            return [blob.as_posix() 
                    for blob in blobs 
                    if blob.is_file() and 
                        (blob.suffix in self.VALID_EXTENSIONS or blob.stem in self.VALID_EXTENSIONS)]
        return []

    def get_blob_stream(self, blob_name:str) -> io.IOBase:
        with open(blob_name, 'rb') as f:
            io_stream = io.BytesIO(f.read())
            return io_stream



"""
File Bin
"""
import time
from .base_bin import BaseBin
import os


class FileBin(BaseBin):

    def __init__(
            self,
            bin_name: str,
            path: str):
        self.path = path
        self.name = bin_name
        os.makedirs(self.path, exist_ok=True)

    def __call__(
            self,
            record: str,
            id_: str = ''):

        os.makedirs(F"{self.path}/{self._date_part()}", exist_ok=True)
        filename = F"{self.path}/{self._date_part()}/{id_}{time.time_ns()}.txt"
        with open(filename, 'w') as file:
            file.write(record)
        return filename


"""
LRU Index

Implements an arbitrary length LRU Index, this is not a LRU cache,
this is the 'brains' of a deduplication engine - where the most
recently seen 'size' items are remembered and older items are
ejected.

This relies on dictionaries being ordered.

#nodoc - don't add to the documentation wiki
"""
from typing import Any, List


class LruIndex(object):

    __slots__ = ("hash_list", "size")

    def __init__(self, size: int = 1000):
        self.hash_list: dict = {}
        self.size = size + 1

    def test(self, item: Any):
        # this offers a minor performance improvement
        hash_list = self.hash_list
        # pop() will remove the item if it's in the dict
        # we return a default of False
        if hash_list.pop(item, False):
            hash_list[item] = True
            return True
        # add the item to the top of the dict
        hash_list[item] = True
        if len(hash_list) == self.size:
            # we want to remove the first item in the dict
            # we could convert to a list, but then we need
            # to create a list, this is much faster and
            # uses less memory
            # deepcode ignore unguarded~next~call: will not error
            hash_list.pop(next(iter(hash_list)))
        return False

    def __call__(self, item: Any):
        return self.test(item)



import os
import logging
from functools import lru_cache
from .add_level import add_logging_level
from .log_formatter import LogFormatter
from .levels import LEVELS_TO_STRING, LEVELS
from .google_cloud_logger import GoogleLogger

LOG_NAME: str = "MABEL"
LOG_FORMAT: str = "{BOLD_CYAN}%(name)s{OFF} | %(levelname)-8s | %(asctime)s | {GREEN}%(funcName)s(){OFF} | {YELLOW}%(filename)s{OFF}:{PURPLE}%(lineno)s{OFF} | %(message)s"


def set_log_name(log_name: str):
    global LOG_NAME
    LOG_NAME = log_name
    get_logger.cache_clear()


@lru_cache(1)
def get_logger() -> logging.Logger:
    """
    Use Python's native logging - we created a named logger so we can make sure
    only the logs related to our jobs are included (other modules also use the
    Python's logging module).
    """

    if GoogleLogger.supported():
        return GoogleLogger()  # type:ignore

    logger = logging.getLogger(LOG_NAME)

    # default is to log WARNING and above
    logger.setLevel(int(os.environ.get("LOGGING_LEVEL", 25)))

    # add the TRACE, AUDIT and ALERT levels to the logger
    if not hasattr(logger, "audit"):
        add_logging_level("AUDIT", LEVELS.AUDIT)
    if not hasattr(logging, "alert"):
        add_logging_level("ALERT", LEVELS.ALERT)

    # override the existing handlers for these levels
    add_logging_level("DEBUG", LEVELS.DEBUG)
    add_logging_level("INFO", LEVELS.INFO)
    add_logging_level("WARNING", LEVELS.WARNING)
    add_logging_level("ERROR", LEVELS.ERROR)

    # configure the logger
    mabel_logging_handler = logging.StreamHandler()
    formatter = LogFormatter(
        logging.Formatter(LOG_FORMAT, datefmt="%Y-%m-%dT%H:%M:%S%z")
    )
    mabel_logging_handler.setFormatter(formatter)
    logger.handlers.clear()
    logger.addHandler(mabel_logging_handler)

    return logger



import os
from ...data.writers.internals.base_inner_writer import BaseInnerWriter
from ...errors import MissingDependencyError
try:
    from google.auth.credentials import AnonymousCredentials  # type:ignore
    from google.cloud import storage  # type:ignore
    google_cloud_storage_installed = True
except ImportError:   # pragma: no cover
    google_cloud_storage_installed = False


class GoogleCloudStorageWriter(BaseInnerWriter):

    def __init__(
            self,
            project: str,
            **kwargs):
        if not google_cloud_storage_installed:  # pragma: no cover
            raise MissingDependencyError("`google-cloud-storage` is missing, please install or include in requirements.txt")

        super().__init__(**kwargs)

        # this means we're testing
        if os.environ.get("STORAGE_EMULATOR_HOST") is not None:
            client = storage.Client(
                credentials=AnonymousCredentials(),
                project=project,
            )
        else:  # pragma: no cover
            client = storage.Client(project=project)
        self.gcs_bucket = client.get_bucket(self.bucket)
        self.filename = self.filename_without_bucket

    def commit(
            self,
            byte_data,
            override_blob_name=None):

        # if we've been given the filename, use that, otherwise get the
        # name from the path builder
        if override_blob_name:
            blob_name = override_blob_name
        else:
            blob_name = self._build_path()

        blob = self.gcs_bucket.blob(blob_name)
        blob.upload_from_string(
                byte_data,
                content_type="application/octet-stream")
        return blob_name



import threading
import tempfile
import os
from typing import Any
from ...formats.json import serialize
from ....logging import get_logger
from ....utils.paths import get_parts
from ....index.index import IndexBuilder
from ....errors import MissingDependencyError


BLOB_SIZE = 32*1024*1024  # about 32 files per gigabyte
BUFFER_SIZE = BLOB_SIZE   # buffer in memory really
SUPPORTED_FORMATS_ALGORITHMS = ('jsonl', 'lzma', 'zstd', 'parquet', 'text')
MAXIMUM_RECORDS = 64000   # needs to be less than 2^16-1


def safe_field_name(field_name):
    """strip all the non-alphanums from a field name"""
    import re
    pattern = re.compile('[^a-zA-Z0-9]+')
    return pattern.sub('', field_name)


class BlobWriter():

    def __init__(
            self,
            *,    # force params to be named
            inner_writer=None,  # type:ignore
            blob_size: int = BLOB_SIZE,
            format: str = 'zstd',
            **kwargs):

        self.indexes = kwargs.get('indexes', [])
        if self.indexes != []:
            get_logger().warning("Index functionality is Alpha - interface and features subject to change.")

        self.format = format
        self.maximum_blob_size = blob_size

        if format not in SUPPORTED_FORMATS_ALGORITHMS:
            raise ValueError(F'Invalid format `{format}`, valid options are {SUPPORTED_FORMATS_ALGORITHMS}')

        kwargs['format'] = format
        self.inner_writer = inner_writer(**kwargs)  # type:ignore
        self._open_blob()

    def append(self, record: dict = {}):
        # serialize the record
        if self.format == 'text':
            serialized = str(record).encode() + b'\n'
        else:
            serialized = serialize(record, as_bytes=True) + b'\n'  # type:ignore

        # add the columns to the index
        for column in self.indexes:
            self.index_builders[column].add(self.records_in_blob, record)

        # the newline isn't counted so add 1 to get the actual length
        # if this write would exceed the blob size, close it so another
        # blob will be created
        self.bytes_in_blob += len(serialized) + 1
        if self.bytes_in_blob > self.maximum_blob_size or self.records_in_blob == MAXIMUM_RECORDS:
            self.commit()
            self._open_blob()

        # write the record to the file
        self.file.write(serialized)
        self.records_in_blob += 1

        return self.records_in_blob

    def commit(self):

        committed_blob_name = ''

        if self.bytes_in_blob > 0:
            with threading.Lock():
                try:
                    self.file.flush()
                    self.file.close()
                except ValueError:
                    pass

                if self.format == "parquet":
                    try:
                        from pyarrow import json as js  # type:ignore
                        import pyarrow.parquet as pq    # type:ignore
                    except ImportError as err:  # pragma: no cover
                        raise MissingDependencyError("`pyarrow` is missing, please install or includein requirements.txt")
                    
                    table = js.read_json(self.file_name)
                    pq.write_table(table, self.file_name + '.parquet', compression='ZSTD')
                    self.file_name += '.parquet'

                with open(self.file_name, 'rb') as f:
                    byte_data = f.read()
                    
                committed_blob_name = self.inner_writer.commit(
                        byte_data=byte_data,
                        override_blob_name=None)

                for column in self.indexes:
                    index = self.index_builders[column].build()
                    
                    bucket, path, stem, suffix = get_parts(committed_blob_name)
                    index_name = bucket + '/' + path + '_SYS.' + stem + '.' + safe_field_name(column) + '.index'
                    
                    index_stream = index._index
                    index_stream.seek(0)
                    committed_index_name = self.inner_writer.commit(
                        byte_data=index_stream.read(),
                        override_blob_name=index_name)

                if 'BACKOUT' in committed_blob_name:
                    get_logger().warning(F"{self.records_in_blob:n} failed records written to BACKOUT partition `{committed_blob_name}`")
                get_logger().debug({
                        "committed_blob": committed_blob_name, 
                        "records": self.records_in_blob, 
                        "raw_bytes": self.bytes_in_blob,
                        "committed_bytes": len(byte_data)
                     })
                try:
                    os.remove(self.file_name)
                except ValueError:
                    pass

                self.bytes_in_blob = 0
                self.file_name = None

        return committed_blob_name


    def _open_blob(self):
        self.file_name = self._create_temp_file_name()
        self.file: Any = open(self.file_name, mode='wb', buffering=BUFFER_SIZE)
        if self.format == 'lzma':
            import lzma
            self.file = lzma.open(self.file, mode='wb')
        if self.format == 'zstd':
            import zstandard  # type:ignore
            self.file = zstandard.open(self.file_name, mode='wb')

        # create index builders
        self.index_builders = {}
        for column in self.indexes:
            self.index_builders[column] = IndexBuilder(column)

        self.bytes_in_blob = 0
        self.records_in_blob = 0

    def __del__(self):
        # this should never be relied on to save data
        self.commit()

    def _create_temp_file_name(self):
        """
        Create a tempfile, get the name and then deletes the tempfile.

        The behaviour of tempfiles is inconsistent between operating systems,
        this helps to ensure consistent behaviour.
        """
        file = tempfile.NamedTemporaryFile(prefix='mabel-', delete=True)
        file_name = file.name
        file.close()
        try:
            os.remove(file_name)
        except OSError:
            pass
        return file_name



import uuid
from .internals.trace_blocks import TraceBlocks
from ..utils import entropy
from ..errors import FlowError
from ..logging import get_logger


class FlowRunner():

    def __init__(self, flow):
        self.flow = flow
        self.cycles = 0

    def __call__(
            self,
            data: dict = {},
            context: dict = None,
            trace_sample_rate: float = 1/1000):
        """
        Create a `run` of a flow and execute with a specific data object.

        Parameters:
            data: dictionary, any (optional)
                The data the flow is to process, opinionated to be a dictionary
                however, any data type is accepted.
            context: dictionary (optional)
                Additional information to support the processing of the data
            trace_sample_rate: float (optional)
                The sample for for to emit trace messages for, default is 
                1/1000.
        """
        if not context:
            context = {} 

        # create a run_id for the message if it doesn't already have one
        if not context.get('run_id'):
            context['run_id'] = str(uuid.uuid4())

        # create a tracer for the message
        if not context.get('execution_trace'):
            context['execution_trace'] = TraceBlocks(uuid=context['run_id'])

        # if trace hasn't been explicitly set - randomly select based on a sample rate
        if not context.get('trace') and trace_sample_rate:
            context['trace'] = entropy.random_range(1, round(1 / trace_sample_rate)) == 1  # nosec

        try:
            # start the flow, walk from the nodes with no incoming links
            for operator_name in self.flow.get_entry_points():
                self.cycles += 1
                self._inner_runner(operator_name=operator_name, data=data, context=context)
                if self.cycles % 1000 == 0:
                    get_logger().debug(F"Executed {self.cycles} cycles of flows.")
        except (Exception, SystemExit) as err:
            # if we have a uncaught failure, make sure it's logged
            get_logger().alert(F'FLOW ABEND - {type(err).__name__} - {err}')  # type:ignore
            raise err

        # if being traced, send the trace to the trace writer
        if context.get('trace', False) and hasattr(self, 'trace_writer'):
            self.trace_writer(context['execution_trace'], id_=str(context.get('run_id')))  #type:ignore


    def _inner_runner(
            self,
            operator_name: str = None,
            data: dict = {},
            context: dict = None):
        """
        Walk the dag/flow by:
        - Getting the function of the current node
        - Execute the function, wrapped in the base class
        - Find the next step by finding outgoing edges
        - Call this method for the next step
        """
        if not context:
            context = {}

        operator = self.flow.get_operator(operator_name)
        if operator is None:
            raise FlowError(F"Invalid Flow - Operator {operator_name} is invalid")
        if not hasattr(operator, "error_writer") and hasattr(self, "error_writer"):
            operator.error_writer = self.error_writer  # type:ignore
        out_going_links = self.flow.get_outgoing_links(operator_name)

        outcome = operator(data, context)

        if outcome:
            if not type(outcome).__name__ in ["generator", "list"]:
                outcome_data, outcome_context = outcome
                outcome = [(outcome_data, outcome_context)]
            for outcome_data, outcome_context in outcome:
                for operator_name in out_going_links:
                    self._inner_runner(operator_name=operator_name, data=outcome_data, context=outcome_context.copy())



"""
Writers are the target specific implementations which commit a temporary file
created by the BlobWriter to different systems, such as the filesystem,
Google Cloud Storage or MinIO.

The primary activity is contained in the .commit() method.
"""
import os
import abc
import uuid
import time
from functools import lru_cache
from ....utils import paths

STEM = "{stem}"

class BaseInnerWriter(abc.ABC):

    def __init__(
            self,
            **kwargs):

        dataset = kwargs.get('dataset')
        self.bucket, path, _, _ = paths.get_parts(dataset)

        if self.bucket == '/':
            self.bucket = ''
        if path == '/':
            path = ''

        self.extension = kwargs.get('extension', '.jsonl')
        if kwargs.get('format', '') in ['lzma', 'zstd', 'parquet']:
            self.extension = self.extension + '.' + kwargs['format']
        if kwargs.get('format') == 'text':
            self.extension = '.txt'

        self.filename = self.bucket + '/' + path + STEM + self.extension
        self.filename_without_bucket = path + STEM + self.extension

    def _build_path(self):
        blob_id = F"{time.time_ns():x}-{self._get_node()}"
        return self.filename.replace(STEM, F'{blob_id}')

    @lru_cache(1)
    def _get_node(self):
        return F"{uuid.getnode():x}-{os.getpid():x}"

    @abc.abstractclassmethod
    def commit(
            self,
            byte_data,
            override_blob_name=None):
        pass



"""
DICT(IONARY) (DATA)SET

A group of functions to assist with handling lists of dictionaries. Intended
to be used for simple operations or pre-filtering and selection before loading
into structures like Pandas.
"""
from typing import Iterator, Any, List, Callable
from .group_by import Groups
from .records import select_record_fields, set_value, order
from ..json import serialize, parse


INNER_JOIN = 'INNER'
LEFT_JOIN = 'LEFT'


def join(
        left: Iterator[dict],
        right: Iterator[dict],
        column: str,
        join_type=INNER_JOIN) -> Iterator[dict]:
    """
    Iterates over the left dictset, matching records fron the right dictset. 
    Dictsets provided to this method are expected to be bounded.

    INNER_JOIN, the default, will discard records unless they appear in both
    tables, LEFT_JOIN will keep all the records fron the left table and add
    records for the right table if a match is found.

    It is recommended that the left table be the larger of the two tables as
    the right table is loaded into memory to perform the matching and look ups.

    Parameters:
        left: iterable of dictionaries
            The 'left' dictset
        right: iterable of dictionaries
            The 'right' dictset
        column: string
            The name of column shared by both dictsets to join on
        join_type: string (optional, default INNER_JOIN)
            The type of join, 'INNER_JOIN' or 'LEFT'

    Yields:
        dictionary
    """
    index = create_index(right, column)
    for record in left:
        value = record.get(column)
        if index.get(value):
            yield {**record, **index[value]}
        elif join_type == LEFT_JOIN:
            yield record


def union(*args) -> Iterator[dict]:
    """
    Append the records from a set of lists together, doesn't ensure columns
    align.

    Parameters:
        args: list of iterables of dictionaries
            The lists of dictionaries to concatenate

    Yields:
        dictionary
    """
    for dictset in args:
        yield from dictset


def create_index(
        dictset: Iterator[dict],
        index_column: str) -> dict:
    """
    Create an index of a file to speed up look-ups, it is expected that the
    value in the index_column is unique but this is not enforced.

    Parameters:
        dictset: iterable of dictionaries
            The dictset to process
        index_column: string
            the column in the dictset to index on

    Returns:
        dictionary
    """
    index = {}
    for record in dictset:
        index_value = record[index_column]
        index[index_value] = record
    return index


def select_from(
        dictset: Iterator[dict],
        columns: List[str] = ['*'],
        where: Callable = None) -> Iterator[dict]:
    """
    Scan a dictset, filtering rows and selecting columns.

    Parameters:
        dictset: iterable of dictionaries
            The dictset to process
        columns: list of strings
            The list of column names to return
        where: callable (optional)
            The function to apply to filter records, we return the rows that 
            evaluate to True, default returns all records

    Yields:
        dictionary
    """
    def _select_columns(dictset, columns):
        for record in dictset:
            record = select_record_fields(record, columns)
            yield record

    if where is not None:
        dictset = filter(where, dictset)
    if columns != ['*']:
        dictset = _select_columns(dictset, columns)
    yield from dictset


def set_column(
        dictset: Iterator[dict],
        column_name: str,
        setter: Callable) -> Iterator[dict]:
    """
    Performs set_value on each row in a set.

    Parameters:
        dictset: iterable of dictionaries
            The dictset to process
        column_name: string
            The column to create or update
        setter: callable or any
            A function or constant to update the column with

    Yields:
        dictionary
    """
    for record in dictset:
        yield set_value(record, column_name, setter)


def drop_duplicates(
        dictset: Iterator[dict],
        cache_size: int = 10000):
    """
    NOTE:
        THIS MAY NOT DO WHAT YOU EXPECT IT TO.

    Removes duplicate records from a dictset, as it able to run against
    an unbounded (infinite) set, it may not fully deduplicate a set. The
    larger the cache_size the more likely the duplication will be correct
    however, this is at the expense of memory.

    Parameters:
        dictset: iterable of dictionaries:
            The dictset to process
        cache_size: integer (optional):
            the number of records to cache, default 10,000

    Yields:
        dictionary
    """
    from ....index.lru_index import LruIndex
    lru = LruIndex(size=cache_size)

    for record in dictset:
        entry = serialize(record)
        if lru.test(entry):
            continue
        yield record


def limit(
        dictset: Iterator[dict],
        limit: int):
    """
    Returns up to 'limit' number of records.

    Parameters:
        dictset: iterable of dictionaries:
            The dictset to process
        limit: integer:
            the maximum number of records to return

    Yields:
        dictionary
    """
    for counter, record in enumerate(dictset):
        if counter == limit:
            break
        yield record


def dictsets_match(
        dictset_1: Iterator[dict],
        dictset_2: Iterator[dict]):
    """
    Tests if two dictsets match regardless of the order of the order of the
    records in the dictset. Return is True if the sets match.
    
    Note that this will exhaust a generator.

    Parameters:
        dictset_1: iterable of dictionaries:
            The first dictset to match
        dictset_2: iterable of dictionaries:
            The second dictset to match

    Returns:
        boolean
    """
    def _hash_set(dictset: Iterator[dict]):
        xor = 0
        for record in dictset:
            entry = serialize(record)  # type:ignore
            _hash = hash(entry)
            xor = xor ^ _hash
        return xor

    return _hash_set(dictset_1) == _hash_set(dictset_2)


def page_dictset(
        dictset: Iterator[dict],
        page_size: int) -> Iterator:
    """
    Enables paging through a dictset by returning a page of records at a time.

    Parameters:
        dictset: iterable of dictionaries:
            The dictset to process
        page_size: integer:
            The number of records per page

    Yields:
        dictionary
    """
    chunk: list = []
    for record in dictset:
        if len(chunk) >= page_size:
            yield chunk
            chunk = [record]
        else:
            chunk.append(record)
    if chunk:
        yield chunk


def sort(
        dictset: Iterator[dict],
        column: str,
        cache_size: int,
        descending: bool = True):
    """
    NOTE:
        THIS DOES NOT SORT THE ENTIRE DICTSET.

    Sorts a dictset by a column. As it able to run an unbounded dataset it
    may not correctly order a set completely. The larger the cache_size the
    more likely the set will be ordered correctly, at the cost of memory.

    This method works best with partially sorted data, for randomized data
    and a small cache, the effect of sorting is poor; for partially sorted
    data, and/or a large cache, the performance is better.

    Note that if this method is placed in a pipeline, it will need to process
    cache_size number of records before it will emit any records.

    Parameters:
        dictset: iterable of dictionaries:
            The dictset to process
        column: string:
            The field to order by
        cache_size: integer:
            The number of records to cache
        descending: boolean (optional):
            Order greatest first, default True 

    Yields:
        dictionary
    """

    def _sort_key(key):
        """
        called like this: _sort_key(key)(row)
        """
        k = key

        def _inner_sort_key(row):
            return row.get(k)

        return _inner_sort_key

    # cache_size is the high water mark, 3/4 is the low water mark. 
    # We fill the cache to the high water mark, sort it and yield 
    # the top 1/4 before filling again.
    # This reduces the number of times we execute the sorted function
    # which is the slowest part of this method.
    # A cache_size of 1000 has a neglible impact on performance, a 
    # cache_size of 50000 introduces a performance hit of about 15%.
    quarter_cache = max(cache_size // 4, 1)
    cache = []
    for record in dictset:
        cache.append(record)
        if len(cache) > cache_size:
            cache = sorted(cache, key=_sort_key(column), reverse=descending) 
            if descending:
                yield from reversed(cache[:quarter_cache])
            else:
                yield from cache[:quarter_cache]
            del cache[:quarter_cache]
    cache = sorted(cache, key=_sort_key(column), reverse=descending) 
    yield from cache


def to_pandas(
        dictset: Iterator[dict]):
    """
    Load an iterable of dictionaries into a pandas dataframe.

    Parameters:
        dictset: iterable of dictionaries:
            The dictset to load

    Returns:
        pandas dataframe
    """
    import pandas  # type:ignore
    return pandas.DataFrame(dictset)


def extract_column(
        dictset: Iterator[dict],
        column: str) -> list:
    """
    Extract the values from column into a list 

    Parameters:
        dictset: iterable of dictionaries:
            The dictset to extract values from
        column: string:
            The name of the column to extract the values of

    Returns:
        list
    """  
    return [record.get(column) for record in dictset]


def group_by(
        dictset: Iterator[dict],
        column: str) -> Groups:
    """
    Create a Groups object 

    Parameters:
        dictset: iterable of dictionaries:
            The dictset to group
        column: string:
            The name of the column to group by

    Returns:
        mabel.formats.Groups
    """  
    return Groups(dictset, column)


def jsonify(
        list_of_json_strings: Iterator[dict]):
    """
    Convert a list of strings to a list of dictionaries

    Parameters:
        list_of_json_strings: iterable of strings:
            The JSON formatted strings to parse to dictionaries

    Yields:
        dictionary
    """  
    return map(parse, list_of_json_strings)  # type:ignore


def pass_thru_counter(
        dictset: Iterator[dict]):
    """
    Count the number of records in a dictset that passes through this function.

    The result is returned as an exception when the set is exhausted.

    Parameters:
        dictset: iterable of dictionaries:
            The dictset to count

    Raises:
        Exception with the number of records in the message
    """
    counter = -1
    for counter, record in enumerate(dictset):
        yield record
    raise Exception(counter + 1)


from ..flows.internals.base_operator import BaseOperator
from ..data import Reader


class ReaderOperator(BaseOperator):

    def __init__(
            self,
            *args,
            **kwargs):
        super().__init__(*args, **kwargs)
        self.reader = Reader(**kwargs)

    def execute(self, data, context):
        for row in self.reader:
            yield row, context



from .end_operator import EndOperator
from .filter_operator import FilterOperator
from .noop_operator import NoOpOperator
from .validator_operator import ValidatorOperator
from .profile_data_operator import ProfileDataOperator
from .reader_operator import ReaderOperator
    


"""
Base for new exception types.

#nodoc - don't add to the documentation wiki
"""


class BaseException(Exception):
    def __call__(self, *args):
        return self.__class__(*(self.args + args))



from setuptools import setup, find_packages  # type:ignore

with open('mabel/version.py', 'r') as v:
    vers = v.read()
exec(vers)  # nosec

with open('README.md', 'r') as rm:
    long_description = rm.read()

setup(
   name='mabel',
   version=__version__,
   description='Python Data Libraries',
   long_description=long_description,
   long_description_content_type="text/markdown",
   maintainer='Joocer',
   packages=find_packages(include=['mabel', 'mabel.*']),
   url="https://github.com/mabel-dev/mabel/",
   install_requires=[
        'ujson',
        'python-dateutil',
        'zstandard',
        'mmh3',
        'pydantic'
   ]
)



import io
import mmh3  # type:ignore
import struct
from operator import itemgetter
from typing import Iterable


UNSET = 65535  # 2^16 - 1
MAX_INDEX = 4294967295  # 2^32 - 1
STRUCT_DEF = "I H H"  # 4 byte unsigned int, 2 byte unsigned int, 2 byte unsigned int
RECORD_SIZE = struct.calcsize(STRUCT_DEF)  # this should be 8

"""
There are overlapping terms because we're traversing a dataset so we can traverse a
dataset. 

Terminology:
    Entry     : a record in the Index
    Location  : the position of the entry in the Index
    Position  : the position of the row in the target file
    Row       : a record in the target file
"""


def safe_field_name(field_name):
    """strip all the non-alphanums from a field name"""
    import re
    pattern = re.compile('[^a-zA-Z0-9]+')
    return pattern.sub('', field_name)


class IndexEntry:
    """
    Python friendly representation of index entries.

    Includes binary translations for reading and writing to the index.
    """

    def to_bin(self) -> bytes:
        """
        Convert a model to _bytes_
        """
        return struct.pack(STRUCT_DEF, self.value, self.location, self.count)

    @staticmethod
    def from_bin(buffer):
        """
        Convert _bytes_ to a model
        """
        value, location, count = struct.unpack(STRUCT_DEF, buffer)
        return IndexEntry(value=value, location=location, count=count)

    def __init__(self, value, location, count):
        self.value = value
        self.location = location
        self.count = count


class Index:
    def __init__(self, index: io.BytesIO):
        """
        A data index which speeds up reading data files.

        The file format is fixed-length binary, the search algorithm is a
        classic binary search.
        """
        if isinstance(index, io.BytesIO):
            self._index = index
            # go to the end of the stream
            index.seek(0, 2)
            # divide the size of the stream by the record size to get the
            # number of entries in the index
            self.size = index.tell() // RECORD_SIZE

    @staticmethod
    def build_index(dictset: Iterable[dict], column_name: str):
        """
        Build an index from a dictset.

        Parameters:
            dictset: iterable of dictionaries
                The dictset to index
            column_name: string
                The name of the index which will be indexed

        Returns:
            io.BytesIO
        """
        # We do this in two-steps
        # 1) Build an intermediate form of the index as a list of entries
        # 2) Conver that intermediate form into a binary index
        builder = IndexBuilder(column_name)
        for position, row in enumerate(dictset):
            builder.add(position, row)
        return builder.build()

    def _get_entry(self, location: int):
        """
        get a specific entry from the index
        """
        if location >= self.size:
            return IndexEntry(value=-1, location=-1, count=-1)
        self._index.seek(RECORD_SIZE * location)
        return IndexEntry.from_bin(self._index.read(RECORD_SIZE))

    def _locate_record(self, value):
        """
        Use a binary search algorithm to search the index
        """
        left, right = 0, self.size
        while left <= right:
            middle = (left + right) >> 1
            entry = self._get_entry(middle)
            if entry.value == value:
                return middle, entry
            elif entry.value > value:
                right = middle - 1
            else:
                left = middle + 1
        return -1, None

    def _inner_search(self, search_term) -> Iterable:
        # hash the value and make fit in a four byte unsinged int
        value = mmh3.hash(f"{search_term}") % MAX_INDEX

        # search for an instance of the value in the index
        location, found_entry = self._locate_record(value)

        # we didn't find the entry
        if not found_entry:
            return []

        # the found_entry is the fastest record to be found, this could
        # be the first, last or middle of the set. The count field tells
        # us how many rows to go back, but not how many forward
        start_location = location - found_entry.count + 1
        end_location = location + 1
        while self._get_entry(end_location).value == value:
            end_location += 1

        # extract the row numbers in the target dataset
        return [
            self._get_entry(loc).location
            for loc in range(start_location, end_location, 1)
        ]

    def search(self, search_term) -> Iterable:
        """
        Search the index for a value. Returns a list of row numbers, if the value is
        not found, the list is empty.
        """
        if not isinstance(search_term, (list, set, tuple)):
            search_term = [search_term]
        result: list = []
        for term in search_term:
            result += self._inner_search(term)
        return set(result)


class IndexBuilder:

    slots = ("column_name", "temporary_index")

    def __init__(self, column_name: str):
        self.column_name: str = column_name
        self.temporary_index: Iterable[dict] = []

    def add(self, position, record):
        if record.get(self.column_name):
            # index lists of items separately
            values = record[self.column_name]
            if not isinstance(values, list):
                values = [values]
            for value in values:
                entry = {
                    "value": mmh3.hash(f"{value}") % MAX_INDEX,
                    "position": position,
                }
                self.temporary_index.append(entry)

    def build(self) -> Index:
        previous_value = None
        index = bytes()
        count: int = 0
        self.temporary_index = sorted(self.temporary_index, key=itemgetter("value"))
        for row in self.temporary_index:
            if row["value"] == previous_value:
                count += 1
            else:
                count = 1
            index += IndexEntry(
                value=row["value"], location=row["position"], count=count
            ).to_bin()
            previous_value = row["value"]
        return Index(io.BytesIO(index))



import os
import re
import hashlib
import logging
import ujson as json  # use ujson rather than orjson for compatibility
from functools import lru_cache
from ..utils.colors import COLORS, colorize
from ..utils.ipython import is_running_from_ipython


# if we find a key which matches these strings, we hash the contents
KEYS_TO_SANITIZE = ["password$", "pwd$", ".*_secret$", ".*_key$"]
COLOR_EXCHANGES = {
    " ALERT    ": "{BOLD_RED} ALERT    {OFF}",
    " ERROR    ": "{RED} ERROR    {OFF}",
    " DEBUG    ": "{GREEN} DEBUG    {OFF}",
    " AUDIT    ": "{YELLOW} AUDIT    {OFF}",
    " WARNING  ": "{CYAN} WARNING  {OFF}",
    " INFO     ": "{BOLD_WHITE} INFO     {OFF}",
}


class LogFormatter(logging.Formatter):
    def __init__(self, orig_formatter):
        """
        Remove sensitive data from records before saving to external logs.
        Note that the value is hashed using (SHA256) and only the first 8
        characters of the hex-encoded hash are presented. This information
        allows values to be traced without disclosing the actual value.

        The Sanitizer can only sanitize dictionaries, it doesn't
        sanitize strings, which could contain sensitive information
        We use the message id as a salt to further protect sensitive
        information.

        Based On: https://github.com/joocer/cronicl/blob/main/cronicl/utils/sanitizer.py
        """
        self.orig_formatter = orig_formatter

    def format(self, record):
        msg = self.orig_formatter.format(record)
        msg = self.sanitize_record(msg)
        if "://" in msg:
            msg = re.sub(r":\/\/(.*?)\@", r"://{BOLD_PURPLE}<redacted>{OFF}", msg)
        return msg

    @lru_cache(1)
    def _can_colorize(self):

        if is_running_from_ipython():
            return True

        colorterm = os.environ.get("COLORTERM", "").lower()
        term = os.environ.get("TERM", "").lower()
        return "yes" in colorterm or "true" in colorterm or "256" in term

    def color_code(self, record):
        if self._can_colorize():
            for k, v in COLOR_EXCHANGES.items():
                if k in record:
                    return record.replace(k, v)
        return record

    def __getattr__(self, attr):
        return getattr(self.orig_formatter, attr)

    def hash_it(self, value_to_hash):
        return hashlib.sha256(value_to_hash.encode()).hexdigest()[:8]

    def sanitize_record(self, record):

        record = self.color_code(record)
        parts = record.split("|")
        json_part = parts.pop()

        try:
            dirty_record = json.loads(json_part)
        except ValueError:
            dirty_record = {"message": json_part.strip()}

        if isinstance(dirty_record, dict):
            clean_record = {}
            for key, value in dirty_record.items():
                if any(
                    [
                        True
                        for expression in KEYS_TO_SANITIZE
                        if re.match(expression, key, re.IGNORECASE)
                    ]
                ):
                    clean_record["{BLUE}" + key + "{OFF}"] = (
                        "{PURPLE}<redacted:" + self.hash_it(str(value)) + ">{OFF}"
                    )
                else:
                    value = re.sub("`([^`]*)`", r"`{YELLOW}\1{GREEN}`", f"{value}")
                    value = re.sub("'([^']*)'", r"'{YELLOW}\1{GREEN}'", f"{value}")
                    clean_record["{BLUE}" + key + "{OFF}"] = (
                        "{GREEN}" + f"{value}" + "{OFF}"
                    )

            parts.append(" " + json.dumps(clean_record))

        record = "|".join(parts)
        return colorize(record, self._can_colorize())



from ..flows.internals.base_operator import BaseOperator

def match_all(data):
    return True

class FilterOperator(BaseOperator):

    def __init__(self, condition=match_all):
        """
        Filters records, returns the record for matching records and returns 'None'
        for non-matching records.

        The Filter Operator takes one configuration item at creation - a _Callable_
        (function) which takes the data as a dictionary as it's only parameter and
        returns `True` to retain the record, or `False` to not pass the record 
        through to the next Operator.

        Example:
            fo = FilterOperator(condition=lambda r: r.get('severity') == 'high')

        Note:
            The condition does not need to be lambda, it can be any _Callable_
            including methods.
        """
        self.condition = condition
        super().__init__()

    def execute(self, data={}, context={}):
        if self.condition(data):
            return data, context
        return None



import io
from ...data.writers.internals.base_inner_writer import BaseInnerWriter
from ...errors import MissingDependencyError
try:
    from minio import Minio  # type:ignore
    minio_installed = True
except ImportError:  # pragma: no cover
    minio_installed = False


class MinIoWriter(BaseInnerWriter):

    def __init__(
            self,
            *,
            end_point: str,
            access_key: str,
            secret_key: str,
            secure: bool = False,
            **kwargs):

        if not minio_installed:  # pragma: no cover
            raise MissingDependencyError("`minio` is missing, please install or include in requirements.txt")
        super().__init__(**kwargs)

        self.client = Minio(end_point, access_key, secret_key, secure=secure)
        self.filename = self.filename_without_bucket

    def commit(
            self,
            byte_data,
            override_blob_name=None):

        # if we've been given the filename, use that, otherwise get the
        # name from the path builder
        if override_blob_name:
            blob_name = override_blob_name
        else:
            blob_name = self._build_path()

        self.client.put_object(
                    self.bucket,
                    blob_name,
                    io.BytesIO(byte_data),
                    len(byte_data))

        return blob_name



"""
A MongoDB Reader

This is a light-weight and prototype MongoDB reader.
"""
import io
from typing import Iterable, Optional, List
from ...data.readers.internals.base_inner_reader import BaseInnerReader
from ...errors import MissingDependencyError
try:
    import pymongo     # type:ignore
    mongo_installed = True
except ImportError:    # pragma: no cover
    mongo_installed = False


class MongoDbReader(BaseInnerReader):

    def __init__(
            self,
            connection_string: str,
            **kwargs):

        if not mongo_installed:  # pragma: no cover
            raise MissingDependencyError("`pymongo` is missing, please install or include in requirements.txt")

        super().__init__(**kwargs)

        connection = pymongo.MongoClient(connection_string)
        self.database = connection[kwargs.get('dataset')] 

        # chunk size affects memory usage
        self.chunk_size: int = kwargs.get('chunk_size', 10)
        self.query: dict = kwargs.get('query', {})

    def get_list_of_blobs(self):
        return self.database.list_collection_names()

    def get_records(
            self,
            blob_name: str,
            rows: Optional[Iterable[int]] = None) -> Iterable[str]:
        collection = self.database[blob_name]  # type:ignore
        chunks = self._iterate_by_chunks(
                collection,
                self.chunk_size,
                0,
                query=self.query)
        for docs in chunks:
            yield from docs

    def _iterate_by_chunks(self, collection, chunksize=1, start_from=0, query={}):
        chunks = range(start_from, collection.find(query).count(), int(chunksize))
        num_chunks = len(chunks)
        for i in range(1,num_chunks+1):
            if i < num_chunks:
                yield collection.find(query)[chunks[i-1]:chunks[i]]
            else:
                yield collection.find(query)[chunks[i-1]:chunks.stop]

    def get_blobs_at_path(self, path):
        # not used but must be present
        pass

    def get_blob_stream(self, blob_name:str) -> io.IOBase:
        # not used but must be present
        pass



from ...flows.internals.base_writer_operator import BaseWriterOperator
from ...data import StreamWriter
from ...adapters.google import GoogleCloudStorageWriter


class GoogleStorageStreamWriterOperator(BaseWriterOperator):

    def __init__(
            self,
            *,
            project:str = None,
            **kwargs):

        kwargs['project'] = project
        super().__init__(
                writer=StreamWriter,
                inner_writer=GoogleCloudStorageWriter,
                **kwargs)
